"""Load configuration"""

from abc import abstractmethod
from pathlib import Path

import xdg_base_dirs
import yaml

from .local_logging import logger
from .project import PROJECT_NAME


class ConfigLoader:
    """Load configuration file"""

    def __init__(self) -> None:
        super().__init__()

        self.lat: str
        self.lng: str
        self.update_days: int

    @abstractmethod
    def load_config(self):
        """Load configuration"""


class YamlConfigLoader(ConfigLoader):
    """Read yaml config file named config.yaml

    Attributes:
        cofnig_path: path of the config.yaml file
    """

    filename = "config.yaml"
    default_filepath = xdg_base_dirs.xdg_config_home() / PROJECT_NAME

    def __init__(self, config_path: Path = default_filepath) -> None:
        super().__init__()

        if not config_path.exists():
            raise RuntimeError(f"{config_path} is not a valid path. Exiting")

        if config_path.is_dir():
            self.config_filepath = config_path / self.filename
        else:
            self.config_filepath = config_path

        logger.info(f"Config file is {self.config_filepath}")

        if not self.config_filepath.exists():
            raise RuntimeError(
                f"No file named {self.filename} found inside {config_path}. Exiting"
            )

    def load_config(self):
        with open(self.config_filepath, "r") as file:
            configuration = yaml.safe_load(file)
        logger.info(configuration)
        super().load_config()
        self.lat = configuration["lat"]
        self.lng = configuration["lng"]
        self.update_days = configuration["update_days"]
        return configuration



"""Transition to sunrise or sunset"""

from datetime import time

import xdg_base_dirs

from .local_logging import logger
from .project import PROJECT_NAME
from .script_runner import ScriptRunner
from .states import State

Default_Cache_dir = xdg_base_dirs.xdg_cache_home()
Default_Cache_subdir = PROJECT_NAME
Default_cache_filename = "cached_state"


class Transitions:
    """Handle transition between State's"""

    last_transition = None

    def calc_next(
        self,
        sunrise_time: time,
        sunset_time: time,
        time_now: time,
    ) -> State:
        """Calculate the next transition by this logic:
        if e.g. it is 1530 now, the sunrise time was 0700, and the sunset time is 1900
        then the next transition is dark

        Args:
            sunrise_time: sunrise time
            time_now: the time now

        Returns:
            The next state calculated
        """
        logger.debug(
            f"Getting next transition with now time: {time_now}, sunrise time: {sunrise_time}, sunset time: {sunset_time}"
        )
        if time_now >= sunrise_time and time_now < sunset_time:
            logger.debug("Returning Dark")
            return State.DARK

        logger.debug("Returning Light")
        return State.LIGHT

    def execute(self, st: State, runner: ScriptRunner):
        runner.run_scripts_in_dir(st)
        self.update_last_transition(st)

    def update_last_transition(self, st: State):
        self.last_transition = st



"""Run scripts inside a dir"""

import os
from pathlib import Path
import subprocess

import xdg_base_dirs
import xdg_base_dirs

from blueblack import project

from .local_logging import logger
from .states import State


class ScriptRunner:
    """Run scripts inside a dir

    Attributes:
        default_filepath: Default directory to run scripts in
        defaults to ~/.config/blueblack/{dark,light}_mode if xdg_config_home is defined
    """

    default_filepath = xdg_base_dirs.xdg_config_home() / project.PROJECT_NAME

    def __init__(self, scripts_path: Path) -> None:
        self.scripts_path = scripts_path

    def get_dir(self, to_mode: State):
        """Return the default dir filepath according to next transition

        Args:
            transition_to: next transition

        Returns:
            Retrurn the complete path
        """
        if to_mode == State.DARK:
            return self.scripts_path / Path("dark_mode")
        elif to_mode == State.LIGHT:
            return self.scripts_path / Path("light_mode")

    def run_scripts_in_dir(self, transition_to: State):
        """Run scripts inside the default dir

        Args:
            transition_to: next transition ["light"|"dark"]
        """
        directory = self.get_dir(transition_to)
        if not directory.exists() or not directory.is_dir():
            logger.error(f"{directory} does not exist. Will not run anything")
            return
        logger.info(f"Running scripts in {directory}")
        for filename in os.listdir(directory):
            f = os.path.join(directory, filename)
            # checking if it is a file
            if os.path.isfile(f):
                if os.access(f, os.X_OK):
                    output = subprocess.run(
                        f, capture_output=True, timeout=10, check=False
                    )
                    if output.returncode != 0:
                        logger.error(
                            f"File {f} returned error code {output.returncode}"
                        )
                else:
                    logger.warn(f"File {f} is not executable")



"""Main file"""

import time
from datetime import UTC, datetime

from blueblack import suntimes_fetcher
from blueblack.config_loading import YamlConfigLoader
from blueblack.local_logging import logger
from blueblack.script_runner import ScriptRunner
from blueblack.states import State
from blueblack.transitions import Transitions

if __name__ == "__main__":
    yaml_config_loader = YamlConfigLoader()
    yaml_config_loader.load_config()

    suntimes_fetcher = suntimes_fetcher.SunTimesFetcherFromApi()
    suntimes_fetcher.setup(yaml_config_loader.lat, yaml_config_loader.lng)
    suntimes = suntimes_fetcher.fetch_sun_times()

    transition = Transitions()

    script_runner = ScriptRunner(ScriptRunner.default_filepath)

    seconds_in_day = 86400
    sleep_interval = 5
    seconds_rem = seconds_in_day * yaml_config_loader.update_days

    while True:
        now_time = datetime.now(UTC).timetz()

        next_transition = transition.calc_next(
            suntimes.sunrise_time, suntimes.sunset_time, now_time
        )

        # Apply opposite here
        if (
            transition.last_transition is None
            or transition.last_transition == next_transition
        ):
            logger.debug(
                f"Skipped a transition or new execution. Will apply opposite of {next_transition}"
            )
            if next_transition == State.DARK:
                transition.execute(State.LIGHT, script_runner)
            else:
                transition.execute(State.DARK, script_runner)

        elif transition.last_transition != next_transition:
            # check if we are past the time for a transition
            if next_transition == State.DARK:
                if now_time > suntimes.sunset_time or now_time < suntimes.sunrise_time:
                    transition.execute(State.DARK, script_runner)
            if next_transition == State.LIGHT:
                if now_time > suntimes.sunrise_time and now_time < suntimes.sunset_time:
                    transition.execute(State.LIGHT, script_runner)

        # check if it is time to update the sunrise, sunset times
        if seconds_rem <= 0:
            suntimes = suntimes_fetcher.fetch_sun_times()
            seconds_rem = seconds_in_day
        else:
            seconds_rem = seconds_rem - sleep_interval

        logger.debug("Sleeping now...")
        time.sleep(sleep_interval)



"""Logging configuration"""

import logging

logger = logging
FORMAT = "[{asctime:<20} {filename:>20}:{lineno:<4} - {funcName:<20}]   {message}"
logging.basicConfig(format=FORMAT, level=logging.DEBUG, style="{")



from datetime import UTC, datetime
import json
from abc import abstractmethod

import requests

from .local_logging import logger
from .sun_times import SunTimes
from jsonschema import validate


class SunTimesFetcher:
    """Fetches sunrise and sunset times in UTC"""

    @abstractmethod
    def fetch_sun_times(self) -> SunTimes:
        """Get the sunsrise and sunset times for your location in UTC
        Returns: Return sunrise"""
        pass


class SunTimesFetcherFromApi(SunTimesFetcher):
    """Get sunrise/sunset times from an API call
    to default https://api.sunrise-sunset.org/json"""

    default_api_uri = "https://api.sunrise-sunset.org/json"
    expected_schema = {
        "type": "object",
        "properties": {
            "status": {"type": "string"},
            "results": {
                "type": "object",
                "properties": {
                    "sunrise": {"type": "string"},
                    "sunset": {"type": "string"},
                },
                "additionalProperties": True,
                "required": ["sunrise", "sunset"],
            },
        },
        "required": ["status", "results"],
        "additionalProperties": False,
    }

    def validate_response(self, response):
        return validate(instance=response, schema=self.expected_schema)

    def setup(self, lat: str, lng: str, endpoint: str = default_api_uri):
        self.endpoint = endpoint
        self.lat = lat
        self.lng = lng

    def run_request(self, lat: str, lng: str) -> SunTimes:
        """Run a get request to the endpoint

        Raises:
            RuntimeError: for not 200 responses
            RuntimeError: for not OK responses from the API

        Returns:
            Returns a dict with keys:
            ["sunrise"]
            ["sunset"]
            They include datetimes for sunrise and sunset datetimes
        """

        payload = {
            "lat": lat,
            "lng": lng,
            "formatted": "0",
        }

        r = requests.get(self.endpoint, payload, timeout=30)

        if r.status_code != 200:
            logger.critical(f"Error code received: {r.status_code}")
            raise RuntimeError(
                f"Cannot get sun times, did not get 200. Got {r.status_code}"
            )

        # It will return a json
        resp_body = json.loads(r.text)
        if resp_body["status"] != "OK":
            logger.critical(f"Error code received: {resp_body["status"]}")
            raise RuntimeError("Cannot get sun times, did not get OK")

        # These times will be in UTC. Need to adjust them for correctness
        logger.info(f"Request return {resp_body}")

        results = resp_body["results"]

        sunr = results["sunrise"]
        suns = results["sunset"]

        logger.debug(f"Sunrise returned: {sunr} - Sunset returned: {suns}")
        return SunTimes(
            datetime.fromisoformat(sunr).timetz(), datetime.fromisoformat(suns).timetz()
        )

    def _run_request(self) -> SunTimes:
        return self.run_request(self.lat, self.lng)

    def fetch_sun_times(self) -> SunTimes:
        return self._run_request()



from datetime import time


class SunTimes:
    def __init__(self, sunrise_time: time, sunset_time: time) -> None:
        self.sunrise_time = sunrise_time
        self.sunset_time = sunset_time

    @classmethod
    def from_json(cls, json_data: dict):
        return cls(sunrise_time=json_data["sunrise"], sunset_time=json_data["sunset"])
